"""
======
Models
======

Classifiers and regressors.

"""

NAME = "Model"

ID = "orange.widgets.model"

DESCRIPTION = "Prediction"

BACKGROUND = "#FAC1D9"

ICON = "icons/Category-Model.svg"

PRIORITY = 4
