"""
========
Evaluate
========

Evaluating models.

"""

NAME = "Evaluate"

ID = "orange.widgets.evaluate"

DESCRIPTION = "Evaluate model performance"

BACKGROUND = "#C3F3F3"

ICON = "icons/Category-Evaluate.svg"

PRIORITY = 5
